@@ -34,20 +34,29 @@ typedef enum
   TPSIP_DEBUG_MEDIA         = 1 << 1,
   TPSIP_DEBUG_IM            = 1 << 2,
   TPSIP_DEBUG_EVENTS        = 1 << 3,
+  TPSIP_DEBUG_SOFIA         = 1 << 4,
 } TpsipDebugFlags;
 
 void tpsip_debug_set_flags_from_env ();
 void tpsip_debug_set_flags (TpsipDebugFlags flags);
 gboolean tpsip_debug_flag_is_set (TpsipDebugFlags flag);
-void tpsip_debug (TpsipDebugFlags flag, const gchar *format, ...)
-    G_GNUC_PRINTF (2, 3);
+void tpsip_log (TpsipDebugFlags flag, GLogLevelFlags level,
+    const gchar *format, ...) G_GNUC_PRINTF (3, 4);
+void tpsip_debug_free (void);
 
 G_END_DECLS
 
 #ifdef DEBUG_FLAG
 
 #define DEBUG(format, ...) \
-  tpsip_debug(DEBUG_FLAG, "%s: " format, G_STRFUNC, ##__VA_ARGS__)
+  tpsip_log(DEBUG_FLAG, G_LOG_LEVEL_DEBUG, "%s: " format, \
+      G_STRFUNC, ##__VA_ARGS__)
+#define WARNING(format, ...) \
+  tpsip_log(DEBUG_FLAG, G_LOG_LEVEL_WARNING, "%s: " format, \
+      G_STRFUNC, ##__VA_ARGS__)
+#define MESSAGE(format, ...) \
+  tpsip_log(DEBUG_FLAG, G_LOG_LEVEL_MESSAGE, "%s: " format, \
+      G_STRFUNC, ##__VA_ARGS__)
 
 /* #define DEBUGGING tpsip_debug_flag_is_set(DEBUG_FLAG) */
 
@@ -58,9 +67,13 @@ G_END_DECLS
 #ifdef DEBUG_FLAG
 
 #define DEBUG(format, ...)
+#define WARNING(format, ...)
+#define MESSAGE(format, ...)
 
 #endif /* DEBUG_FLAG */
 
+#define tpsip_debug_free() G_STMT_START { } G_STMT_END
+
 #endif /* ENABLE_DEBUG */
 
 #endif /* __DEBUG_H__ */
