@@ -366,8 +366,6 @@ tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
 
   channel = new_media_channel (fac, handle, handle, channel_flags);
 
-  tpsip_media_channel_receive_invite (channel, ev->nua_handle);
-
   tp_handle_unref (contact_repo, handle);
 
   /* We delay emission of NewChannel(s) until we have the data on
@@ -375,6 +373,8 @@ tpsip_nua_i_invite_cb (TpBaseConnection    *conn,
   g_signal_connect (channel, "incoming-call",
       G_CALLBACK (incoming_call_cb), fac);
 
+  tpsip_media_channel_attach_to_nua_handle (channel, ev->nua_handle);
+
   return TRUE;
 }
 
@@ -458,15 +458,15 @@ static const gchar * const named_channel_allowed_properties[] = {
     NULL
 };
 
-/* not advertised in foreach_channel_class - can only be requested with
+/* not advertised in type_foreach_channel_class - can only be requested with
  * RequestChannel, not with CreateChannel/EnsureChannel */
 static const gchar * const anon_channel_allowed_properties[] = {
     NULL
 };
 
 static void
-tpsip_media_factory_foreach_channel_class (TpChannelManager *manager,
-    TpChannelManagerChannelClassFunc func,
+tpsip_media_factory_type_foreach_channel_class (GType type,
+    TpChannelManagerTypeChannelClassFunc func,
     gpointer user_data)
 {
   GHashTable *table = g_hash_table_new_full (g_str_hash, g_str_equal,
@@ -483,7 +483,7 @@ tpsip_media_factory_foreach_channel_class (TpChannelManager *manager,
       handle_type_value);
 
   g_value_set_uint (handle_type_value, TP_HANDLE_TYPE_CONTACT);
-  func (manager, table, named_channel_allowed_properties, user_data);
+  func (type, table, named_channel_allowed_properties, user_data);
 
   g_hash_table_destroy (table);
 }
@@ -695,7 +695,8 @@ channel_manager_iface_init (gpointer g_iface,
   TpChannelManagerIface *iface = g_iface;
 
   iface->foreach_channel = tpsip_media_factory_foreach_channel;
-  iface->foreach_channel_class = tpsip_media_factory_foreach_channel_class;
+  iface->type_foreach_channel_class =
+    tpsip_media_factory_type_foreach_channel_class;
   iface->request_channel = tpsip_media_factory_request_channel;
   iface->create_channel = tpsip_media_factory_create_channel;
   iface->ensure_channel = tpsip_media_factory_ensure_channel;
