@@ -27,8 +27,10 @@
 #include <tpsip/sofia-decls.h>
 #include <sofia-sip/sresolv.h>
 
+#include <telepathy-glib/simple-password-manager.h>
+
 #ifdef HAVE_LIBIPHB
-#include <iphbd/libiphb.h>
+#include <libiphb.h>
 #endif
 
 struct _TpsipConnectionPrivate
@@ -51,6 +53,7 @@ struct _TpsipConnectionPrivate
   gchar *registrar_realm;
 
   TpsipMediaFactory *media_factory;
+  TpSimplePasswordManager *password_manager;
 
   gchar *address;
   gchar *auth_user;
@@ -69,7 +72,6 @@ struct _TpsipConnectionPrivate
   gboolean loose_routing;
   gboolean discover_binding;
   gboolean immutable_streams;
-  gboolean ignore_tls_errors;
 
   gboolean keepalive_interval_specified;
 
