@@ -49,7 +49,6 @@
 #include "sip-connection-sofia.h"
 
 #include <sofia-sip/msg_header.h>
-#include <sofia-sip/tport_tag.h>
 
 #define DEBUG_FLAG TPSIP_DEBUG_CONNECTION
 #include "debug.h"
@@ -103,7 +102,6 @@ enum
   PROP_LOCAL_PORT,         /**< Local port for SIP (normally not needed, chosen by stack) */
   PROP_EXTRA_AUTH_USER,	   /**< User name to use for extra authentication challenges */
   PROP_EXTRA_AUTH_PASSWORD,/**< Password to use for extra authentication challenges */
-  PROP_IGNORE_TLS_ERRORS,  /**< If true, TLS errors will be ignored */
   PROP_SOFIA_ROOT,         /**< Event root pointer from the Sofia-SIP stack */
   LAST_PROPERTY
 };
@@ -162,6 +160,10 @@ tpsip_connection_create_channel_managers (TpBaseConnection *conn)
         "connection", self, NULL);
   g_ptr_array_add (channel_managers, priv->media_factory);
 
+  priv->password_manager = tp_simple_password_manager_new (
+      conn);
+  g_ptr_array_add (channel_managers, priv->password_manager);
+
   return channel_managers;
 }
 
@@ -305,9 +307,6 @@ tpsip_connection_set_property (GObject      *object,
     priv->extra_auth_password =  g_value_dup_string (value);
     break;
   }
-  case PROP_IGNORE_TLS_ERRORS:
-    priv->ignore_tls_errors = g_value_get_boolean (value);
-    break;
   case PROP_SOFIA_ROOT: {
     priv->sofia_root = g_value_get_pointer (value);
     break;
@@ -397,9 +396,6 @@ tpsip_connection_get_property (GObject      *object,
     g_value_set_uint (value, priv->local_port);
     break;
   }
-  case PROP_IGNORE_TLS_ERRORS:
-    g_value_set_boolean (value, priv->ignore_tls_errors);
-    break;
   case PROP_SOFIA_ROOT: {
     g_value_set_pointer (value, priv->sofia_root);
     break;
@@ -430,15 +426,22 @@ static void tpsip_connection_shut_down (TpBaseConnection *base);
 static gboolean tpsip_connection_start_connecting (TpBaseConnection *base,
     GError **error);
 
+static const gchar *interfaces_always_present[] = {
+    TP_IFACE_CONNECTION_INTERFACE_REQUESTS,
+    TP_IFACE_CONNECTION_INTERFACE_CONTACTS,
+    TP_IFACE_CONNECTION_INTERFACE_ALIASING,
+    NULL };
+
+const gchar **
+tpsip_connection_get_implemented_interfaces (void)
+{
+  /* we don't have any conditionally-implemented interfaces */
+  return interfaces_always_present;
+}
+
 static void
 tpsip_connection_class_init (TpsipConnectionClass *klass)
 {
-  static const gchar *interfaces_always_present[] = {
-      TP_IFACE_CONNECTION_INTERFACE_REQUESTS,
-      TP_IFACE_CONNECTION_INTERFACE_CONTACTS,
-      TP_IFACE_CONNECTION_INTERFACE_ALIASING,
-      NULL };
-
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   TpBaseConnectionClass *base_class =
     (TpBaseConnectionClass *)klass;
@@ -531,7 +534,7 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
       "Interval between keepalive probes in seconds "
       "(0 = disabled, unset = use a default interval)",
       0, G_MAXUINT32, 0,
-      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_KEEPALIVE_INTERVAL);
 
   param_spec = g_param_spec_boolean ("discover-binding",
@@ -592,12 +595,6 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
       G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   INST_PROP(PROP_EXTRA_AUTH_PASSWORD);
 
-  param_spec = g_param_spec_boolean ("ignore-tls-errors", "Ignore TLS errors",
-      "If true, the TLS verification errors will be ignored",
-      FALSE,
-      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
-  INST_PROP(PROP_IGNORE_TLS_ERRORS);
-
 #undef INST_PROP
 
   tp_dbus_properties_mixin_class_init (object_class,
@@ -607,6 +604,54 @@ tpsip_connection_class_init (TpsipConnectionClass *klass)
       G_STRUCT_OFFSET (TpsipConnectionClass, contacts_class));
 }
 
+typedef struct {
+  TpsipConnection* self;
+  nua_handle_t *nh;
+  gchar *method;
+  gchar *realm;
+  gchar *user;
+} PrivHandleAuthData;
+
+static PrivHandleAuthData *
+priv_handle_auth_data_new (TpsipConnection* self,
+                           nua_handle_t *nh,
+                           const gchar *method,
+                           const gchar *realm,
+                           const gchar *user)
+{
+  PrivHandleAuthData *data = g_slice_new (PrivHandleAuthData);
+
+  data->self = g_object_ref (self);
+  data->nh = nua_handle_ref (nh);
+  data->method = g_strdup (method);
+  data->realm = g_strdup (realm);
+  data->user = g_strdup (user);
+
+  return data;
+}
+
+static void
+priv_handle_auth_data_free (PrivHandleAuthData *data)
+{
+  g_object_unref (data->self);
+  nua_handle_unref (data->nh);
+  g_free (data->method);
+  g_free (data->realm);
+  g_free (data->user);
+
+  g_slice_free (PrivHandleAuthData, data);
+}
+
+static void priv_password_manager_prompt_cb (GObject *source_object,
+                                             GAsyncResult *result,
+                                             gpointer user_data);
+static void priv_handle_auth_continue (TpsipConnection* self,
+                                       nua_handle_t *nh,
+                                       const gchar *method,
+                                       const gchar *realm,
+                                       const gchar *user,
+                                       const gchar *password);
+
 static gboolean
 priv_handle_auth (TpsipConnection* self,
                   int status,
@@ -621,7 +666,6 @@ priv_handle_auth (TpsipConnection* self,
   const char *realm = NULL;
   const char *user =  NULL;
   const char *password =  NULL;
-  gchar *auth = NULL;
 
   if (status != 401 && status != 407)
     return FALSE;
@@ -649,6 +693,12 @@ priv_handle_auth (TpsipConnection* self,
       return FALSE;
     }
 
+  if (method == NULL)
+    {
+      WARNING ("no method presented for authentication");
+      return FALSE;
+    }
+
   /* step: determine which set of credentials to use */
   if (home_realm)
     {
@@ -682,6 +732,9 @@ priv_handle_auth (TpsipConnection* self,
         /* fall back to the main username */
         user = priv->auth_user;
       password = priv->extra_auth_password;
+      if (password == NULL)
+        /* note that this prevents asking the user for a password */
+        password = "";
 
       DEBUG("using the extra auth credentials");
     }
@@ -692,37 +745,101 @@ priv_handle_auth (TpsipConnection* self,
       if (sipfrom && sipfrom->a_url[0].url_user)
         /* or use the userpart in "From" header */
         user = sipfrom->a_url[0].url_user;
+      else
+        return FALSE;
     }
 
   if (password == NULL)
-    password = "";
+    {
+      PrivHandleAuthData *data = NULL;
+
+      DEBUG("asking the user for a password.");
+      data = priv_handle_auth_data_new (self, nh, method, realm,
+                                        user);
+      tp_simple_password_manager_prompt_async (priv->password_manager,
+          priv_password_manager_prompt_cb, data);
+      /* Promise that we'll handle it eventually, even if we end up just
+       * handling it with a blank password. */
+      return TRUE;
+    }
+
+  priv_handle_auth_continue (self, nh, method, realm,
+                             user, password);
+  return TRUE;
+}
+
+static void
+priv_password_manager_prompt_cb (GObject *source_object,
+                             GAsyncResult *result,
+                             gpointer user_data)
+{
+  PrivHandleAuthData *data = user_data;
+  GError *error = NULL;
+  const GString *password_string;
+  const gchar *password;
+
+  password_string = tp_simple_password_manager_prompt_finish (
+      TP_SIMPLE_PASSWORD_MANAGER (source_object), result, &error);
+
+  if (error != NULL)
+    {
+      /* we promised to handle the auth challenge in priv_handle_auth() by
+       * returning TRUE, so we need to handle it anyway, even if it means
+       * doing it with a blank password.
+       */
+      DEBUG ("Auth channel failed: %s. Using blank password.", error->message);
+
+      password = "";
+
+      g_error_free (error);
+    }
+  else
+    {
+      TpsipConnectionPrivate *priv =
+          TPSIP_CONNECTION_GET_PRIVATE (data->self);
+
+      password = password_string->str;
+      /* also save it for later. */
+      g_free (priv->password);
+      priv->password = g_strdup (password);
+    }
+
+  priv_handle_auth_continue (data->self, data->nh, data->method, data->realm,
+                             data->user, password);
+
+  priv_handle_auth_data_free (data);
+}
+
+static void
+priv_handle_auth_continue (TpsipConnection* self,
+    nua_handle_t *nh,
+    const gchar *method,
+    const gchar *realm,
+    const gchar *user,
+    const gchar *password)
+{
+  gchar *auth = NULL;
 
   /* step: if all info is available, create an authorization response */
   g_assert (realm != NULL);
-  if (user && method) {
-    if (realm[0] == '"')
-      auth = g_strdup_printf ("%s:%s:%s:%s",
-                              method, realm, user, password);
-    else
-      auth = g_strdup_printf ("%s:\"%s\":%s:%s",
-                              method, realm, user, password);
+  g_assert (method != NULL);
+  g_assert (user != NULL);
+  g_assert (password != NULL);
 
-    DEBUG("%s authenticating user='%s' realm=%s",
-          wa ? "server" : "proxy", user, realm);
-  }
+  if (realm[0] == '"')
+    auth = g_strdup_printf ("%s:%s:%s:%s",
+                            method, realm, user, password);
+  else
+    auth = g_strdup_printf ("%s:\"%s\":%s:%s",
+                            method, realm, user, password);
 
-  if (auth == NULL)
-    {
-      WARNING ("authentication data are incomplete");
-      return FALSE;
-    }
+  DEBUG ("%s-authenticating user='%s' realm=%s",
+         method, user, realm);
 
   /* step: authenticate */
   nua_authenticate(nh, NUTAG_AUTH(auth), TAG_END());
 
   g_free (auth);
-
-  return TRUE;
 }
 
 static gboolean
@@ -787,7 +904,7 @@ tpsip_connection_nua_r_register_cb (TpsipConnection     *self,
           if (base->status != TP_CONNECTION_STATUS_CONNECTING)
             return TRUE;
 
-          DEBUG("succesfully registered to the network");
+          DEBUG("successfully registered to the network");
           conn_status = TP_CONNECTION_STATUS_CONNECTED;
           reason = TP_CONNECTION_STATUS_REASON_REQUESTED;
 
@@ -943,8 +1060,6 @@ tpsip_connection_start_connecting (TpBaseConnection *base,
       NUTAG_AUTOANSWER(0),
       NUTAG_APPL_METHOD("MESSAGE"),
       SIPTAG_ALLOW_STR("INVITE, ACK, BYE, CANCEL, OPTIONS, PRACK, MESSAGE, UPDATE"),
-      TAG_IF(!priv->ignore_tls_errors,
-             TPTAG_TLS_VERIFY_POLICY(TPTLS_VERIFY_ALL)),
       TAG_NULL());
   if (priv->sofia_nua == NULL)
     {
@@ -959,10 +1074,10 @@ tpsip_connection_start_connecting (TpBaseConnection *base,
   tpsip_conn_update_nua_keepalive_interval (self);
   tpsip_conn_update_nua_contact_features (self);
 
-  if (priv->discover_stun)
-    tpsip_conn_discover_stun_server (self);
-  else if (priv->stun_host != NULL)
+  if (priv->stun_host != NULL)
     tpsip_conn_resolv_stun_server (self, priv->stun_host);
+  else if (priv->discover_stun)
+    tpsip_conn_discover_stun_server (self);
 
   DEBUG("initialized a Sofia-SIP NUA at address %p", priv->sofia_nua);
 
