@@ -191,7 +191,7 @@ tpsip_media_channel_constructed (GObject *obj)
       G_OBJECT_CLASS (tpsip_media_channel_parent_class);
   TpDBusDaemon *bus;
   TpHandleRepoIface *contact_repo;
-  TpIntSet *set;
+  TpIntSet *add;
 
   if (parent_object_class->constructed != NULL)
     parent_object_class->constructed (obj);
@@ -219,16 +219,17 @@ tpsip_media_channel_constructed (GObject *obj)
   g_assert (priv->initiator != 0);
   tp_handle_ref (contact_repo, priv->initiator);
 
-  set = tp_intset_new ();
-  tp_intset_add (set, priv->initiator);
+  add = tp_intset_new_containing (priv->initiator);
+  tp_group_mixin_change_members (obj, "", add, NULL, NULL, NULL, 0, 0);
+  tp_intset_destroy (add);
 
-  tp_group_mixin_change_members (obj, "", set, NULL, NULL, NULL, 0, 0);
-
-  tp_intset_destroy (set);
-
-  /* Allow member adding; also, we implement the 0.17.6 properties */
+  /* We start off with lots of flags, and then delete them as we work out what
+   * kind of channel we are, rather than trying to track what we need to
+   * add/remove over time. We should always have the right flags before we are
+   * advertised on the bus. */
   tp_group_mixin_change_flags (obj,
-      TP_CHANNEL_GROUP_FLAG_CAN_ADD | TP_CHANNEL_GROUP_FLAG_PROPERTIES, 0);
+      TP_CHANNEL_GROUP_FLAG_CAN_ADD | TP_CHANNEL_GROUP_FLAG_CAN_REMOVE |
+      TP_CHANNEL_GROUP_FLAG_CAN_RESCIND | TP_CHANNEL_GROUP_FLAG_PROPERTIES, 0);
 }
 
 static void tpsip_media_channel_dispose (GObject *object);
@@ -747,7 +748,14 @@ static void
 tpsip_media_channel_get_handle (TpSvcChannel *iface,
                               DBusGMethodInvocation *context)
 {
-  tp_svc_channel_return_from_get_handle (context, 0, 0);
+  TpsipMediaChannel *self = TPSIP_MEDIA_CHANNEL (iface);
+  TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
+
+  if (priv->handle != 0)
+    tp_svc_channel_return_from_get_handle (context, TP_HANDLE_TYPE_CONTACT,
+          priv->handle);
+  else
+    tp_svc_channel_return_from_get_handle (context, TP_HANDLE_TYPE_NONE, 0);
 }
 
 /**
@@ -1032,6 +1040,11 @@ tpsip_media_channel_create_initial_streams (TpsipMediaChannel *self)
 
   g_assert (priv->initiator != priv->handle);
 
+  /* RequestChannel(None, 0) => channel is anonymous:
+   * caller uses RequestStreams to set the peer and start the call. */
+  if (priv->handle == 0)
+    return;
+
   priv_outbound_call (self, priv->handle);
 
   g_assert (priv->session != NULL);
@@ -1049,13 +1062,15 @@ tpsip_media_channel_create_initial_streams (TpsipMediaChannel *self)
         TRUE);
 }
 
-/**
- * Handle an incoming INVITE, normally called just after the channel
- * has been created with initiator handle of the sender.
+/*
+ * Handles an incoming call, called shortly after the channel
+ * has been created with initiator handle of the sender, when remote SDP
+ * session data are reported by the NUA stack.
  */
-void
-tpsip_media_channel_receive_invite (TpsipMediaChannel *self,
-                                    nua_handle_t *nh)
+static void
+tpsip_media_channel_handle_incoming_call (TpsipMediaChannel *self,
+                                          nua_handle_t *nh,
+                                          const sdp_session_t *sdp)
 {
   TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
   TpBaseConnection *conn = TP_BASE_CONNECTION (priv->conn);
@@ -1063,8 +1078,36 @@ tpsip_media_channel_receive_invite (TpsipMediaChannel *self,
   g_assert (priv->initiator != conn->self_handle);
   g_assert (priv->session == NULL);
 
-  /* Start the local stream-engine; once the local
-   * media are ready, reply with nua_respond() */
+  if (sdp != NULL)
+    {
+      /* Get the initial media properties from the session offer */
+      const sdp_media_t *media;
+  
+      for (media = sdp->sdp_media; media != NULL; media = media->m_next)
+        {
+          if (media->m_rejected || media->m_port == 0)
+            continue;
+  
+          switch (media->m_type)
+            {
+            case sdp_media_audio:
+              priv->initial_audio = TRUE;
+              DEBUG("has initial audio");
+              break;
+            case sdp_media_video:
+              priv->initial_video = TRUE;
+              DEBUG("has initial video");
+              break;
+            default:
+              break;
+            }
+        }
+    }
+
+  /* Tell the factory to emit NewChannel(s) */
+  g_signal_emit (self, signals[SIG_INCOMING_CALL], 0);
+
+  /* Offer the session handler to the client */
   priv_create_session (self, nh, priv->initiator);
 
   g_assert (priv->session != NULL);
@@ -1105,7 +1148,7 @@ tpsip_media_channel_peer_error (TpsipMediaChannel *self,
                                 const char* message)
 {
   TpGroupMixin *mixin = TP_GROUP_MIXIN (self);
-  TpIntSet *set;
+  TpIntSet *remove;
   guint reason = TP_CHANNEL_GROUP_CHANGE_REASON_ERROR;
 
   switch (status)
@@ -1142,12 +1185,12 @@ tpsip_media_channel_peer_error (TpsipMediaChannel *self,
   if (message == NULL || !g_utf8_validate (message, -1, NULL))
     message = "";
 
-  set = tp_intset_new ();
-  tp_intset_add (set, peer);
-  tp_intset_add (set, mixin->self_handle);
+  remove = tp_intset_new ();
+  tp_intset_add (remove, peer);
+  tp_intset_add (remove, mixin->self_handle);
   tp_group_mixin_change_members ((GObject *)self, message,
-      NULL, set, NULL, NULL, peer, reason);
-  tp_intset_destroy (set);
+      NULL, remove, NULL, NULL, peer, reason);
+  tp_intset_destroy (remove);
 }
 
 guint
@@ -1191,26 +1234,21 @@ priv_nua_i_bye_cb (TpsipMediaChannel *self,
 {
   TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
   TpGroupMixin *mixin = TP_GROUP_MIXIN (self);
-  TpIntSet *set;
+  TpIntSet *remove;
   TpHandle peer;
 
   g_return_val_if_fail (priv->session != NULL, FALSE);
 
   peer = tpsip_media_session_get_peer (priv->session);
-  set = tp_intset_new ();
-  tp_intset_add (set, peer);
-  tp_intset_add (set, mixin->self_handle);
+  remove = tp_intset_new ();
+  tp_intset_add (remove, peer);
+  tp_intset_add (remove, mixin->self_handle);
 
-  tp_group_mixin_change_members ((GObject *) self,
-                                 "",
-                                 NULL, /* add */
-                                 set,  /* remove */
-                                 NULL,
-                                 NULL,
-                                 peer,
-                                 TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+  tp_group_mixin_change_members ((GObject *) self, "",
+     NULL, remove, NULL, NULL,
+     peer, TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
 
-  tp_intset_destroy (set);
+  tp_intset_destroy (remove);
 
   return TRUE;
 }
@@ -1223,7 +1261,7 @@ priv_nua_i_cancel_cb (TpsipMediaChannel *self,
 {
   TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
   TpGroupMixin *mixin = TP_GROUP_MIXIN (self);
-  TpIntSet *set;
+  TpIntSet *remove;
   TpHandle actor = 0;
   TpHandle peer;
   const sip_reason_t *reason;
@@ -1267,52 +1305,19 @@ priv_nua_i_cancel_cb (TpsipMediaChannel *self,
   if (message == NULL || !g_utf8_validate (message, -1, NULL))
     message = "";
 
-  set = tp_intset_new ();
-  tp_intset_add (set, peer);
-  tp_intset_add (set, mixin->self_handle);
+  remove = tp_intset_new ();
+  tp_intset_add (remove, peer);
+  tp_intset_add (remove, mixin->self_handle);
 
-  tp_group_mixin_change_members ((GObject *) self,
-                                 message,
-                                 NULL, /* add */
-                                 set,  /* remove */
-                                 NULL,
-                                 NULL,
-                                 actor,
-                                 TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+  tp_group_mixin_change_members ((GObject *) self, message,
+      NULL, remove, NULL, NULL,
+      actor, TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
 
-  tp_intset_destroy (set);
+  tp_intset_destroy (remove);
 
   return TRUE;
 }
 
-static void
-priv_initial_media_properties_from_sdp (TpsipMediaChannel *self,
-                                        const sdp_session_t *sdp)
-{
-  TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
-  const sdp_media_t *media;
-
-  for (media = sdp->sdp_media; media != NULL; media = media->m_next)
-    {
-      if (media->m_rejected || media->m_port == 0)
-        continue;
-
-      switch (media->m_type)
-        {
-        case sdp_media_audio:
-          priv->initial_audio = TRUE;
-          DEBUG("has initial audio");
-          break;
-        case sdp_media_video:
-          priv->initial_video = TRUE;
-          DEBUG("has initial video");
-          break;
-        default:
-          break;
-        }
-    }
-}
-
 static gboolean
 priv_nua_i_state_cb (TpsipMediaChannel *self,
                      const TpsipNuaEvent  *ev,
@@ -1327,8 +1332,6 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
   gint status = ev->status;
   TpHandle peer;
 
-  g_return_val_if_fail (priv->session != NULL, FALSE);
-
   tl_gets(tags,
           NUTAG_CALLSTATE_REF(ss_state),
           NUTAG_OFFER_RECV_REF(offer_recv),
@@ -1336,6 +1339,18 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
           SOATAG_REMOTE_SDP_REF(r_sdp),
           TAG_END());
 
+  DEBUG("call with handle %p is %s", ev->nua_handle, nua_callstate_name (ss_state));
+
+  if (ss_state == nua_callstate_received && priv->session == NULL)
+    {
+      /* We get the session data for initial media properties with this event;
+       * initialize the session before we can create any streams below.
+       */
+      tpsip_media_channel_handle_incoming_call (self, ev->nua_handle, r_sdp);
+    }
+
+  g_return_val_if_fail (priv->session != NULL, FALSE);
+
   if (r_sdp)
     {
       g_return_val_if_fail (answer_recv || offer_recv, FALSE);
@@ -1348,18 +1363,8 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
 
   peer = tpsip_media_session_get_peer (priv->session);
 
-  DEBUG("call with handle %p is %s", ev->nua_handle, nua_callstate_name (ss_state));
-
   switch ((enum nua_callstate)ss_state)
     {
-    case nua_callstate_received:
-      if (tpsip_media_session_get_state (priv->session)
-          != TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED)
-        break;
-      if (r_sdp != NULL)
-        priv_initial_media_properties_from_sdp (self, r_sdp);
-      g_signal_emit (self, signals[SIG_INCOMING_CALL], 0);
-      break;
     case nua_callstate_proceeding:
       switch (status)
         {
@@ -1389,7 +1394,22 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
           TPSIP_CHANNEL_CALL_STATE_PROCEEDING_MASK);
 
       if (status < 300)
-        tpsip_media_session_accept (priv->session);
+        {
+          TpIntSet *add = tp_intset_new_containing (peer);
+
+          tp_group_mixin_change_members ((GObject *) self,
+                                         "",
+                                         add,  /* add */
+                                         NULL, /* remove */
+                                         NULL,
+                                         NULL,
+                                         peer,
+                                         TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+
+          tp_intset_destroy (add);
+
+          tpsip_media_session_accept (priv->session);
+        }
       else if (status == 491)
         tpsip_media_session_resolve_glare (priv->session);
       else
@@ -1443,12 +1463,10 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
   switch (state)
     {
     case TPSIP_MEDIA_SESSION_STATE_INVITE_SENT:
-      set = tp_intset_new ();
-
       g_assert (priv->initiator == self_handle);
 
       /* add the peer to remote pending */
-      tp_intset_add (set, peer);
+      set = tp_intset_new_containing (peer);
       tp_group_mixin_change_members ((GObject *)channel,
                                      "",
                                      NULL,    /* add */
@@ -1458,18 +1476,15 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
                                      self_handle,        /* actor */
                                      TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
 
-      /* update flags: allow removal and rescinding, no more adding */
-      tp_group_mixin_change_flags ((GObject *)channel,
-          TP_CHANNEL_GROUP_FLAG_CAN_REMOVE | TP_CHANNEL_GROUP_FLAG_CAN_RESCIND,
+      /* update flags: no more adding */
+      tp_group_mixin_change_flags ((GObject *)channel, 0,
           TP_CHANNEL_GROUP_FLAG_CAN_ADD);
 
       break;
 
     case TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED:
-      set = tp_intset_new ();
-
       /* add ourself to local pending */
-      tp_intset_add (set, self_handle);
+      set = tp_intset_new_containing (self_handle);
       tp_group_mixin_change_members ((GObject *) channel, "",
                                      NULL,          /* add */
                                      NULL,          /* remove */
@@ -1478,10 +1493,15 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
                                      priv->initiator, /* actor */
                                      TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
 
-      /* No adding more members to the incoming call, removing is OK */
-      tp_group_mixin_change_flags ((GObject *) channel,
-          TP_CHANNEL_GROUP_FLAG_CAN_REMOVE,
-          TP_CHANNEL_GROUP_FLAG_CAN_ADD);
+      /* No adding more members to the incoming call. Therefore also not
+       * possible to add anyone to remote-pending, so rescinding would make
+       * utterly no sense. We also disallow removing the remote peer if
+       * we are not the initiator, so disallow that too.
+       * Removing yourself to end the call is not represented by group flags.
+       */
+      tp_group_mixin_change_flags ((GObject *) channel, 0,
+          TP_CHANNEL_GROUP_FLAG_CAN_ADD | TP_CHANNEL_GROUP_FLAG_CAN_REMOVE |
+          TP_CHANNEL_GROUP_FLAG_CAN_RESCIND);
 
       break;
 
@@ -1491,10 +1511,8 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
           if (!tp_handle_set_is_member (mixin->remote_pending, peer))
             break; /* no-op */
 
-          set = tp_intset_new ();
-
           /* the peer has promoted itself to members */
-          tp_intset_add (set, peer);
+          set = tp_intset_new_containing (peer);
           tp_group_mixin_change_members ((GObject *)channel, "",
                                          set,     /* add */
                                          NULL,    /* remove */
@@ -1507,10 +1525,8 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
           if (!tp_handle_set_is_member (mixin->local_pending, self_handle))
             break; /* no-op */
 
-          set = tp_intset_new ();
-
           /* promote ourselves to members */
-          tp_intset_add (set, self_handle);
+          set = tp_intset_new_containing (self_handle);
           tp_group_mixin_change_members ((GObject *)channel, "",
                                          set,     /* add */
                                          NULL,    /* remove */
@@ -1519,11 +1535,12 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
                                          self_handle, 0);
         }
 
-      /* update flags: allow removal, deny adding and rescinding */
-      tp_group_mixin_change_flags ((GObject *)channel,
-          TP_CHANNEL_GROUP_FLAG_CAN_REMOVE,
-          TP_CHANNEL_GROUP_FLAG_CAN_ADD
-          | TP_CHANNEL_GROUP_FLAG_CAN_RESCIND);
+      /* update flags: deny adding and rescinding. Removing the remote peer is
+       * still allowed.
+       * Removing yourself to end the call is not represented by group flags.
+       */
+      tp_group_mixin_change_flags ((GObject *)channel, 0,
+          TP_CHANNEL_GROUP_FLAG_CAN_ADD | TP_CHANNEL_GROUP_FLAG_CAN_RESCIND);
 
       break;
 
@@ -1551,8 +1568,9 @@ static void priv_session_state_changed_cb (TpsipMediaSession *session,
     tp_intset_destroy (set);
 }
 
-static void
-priv_connect_nua_handlers (TpsipMediaChannel *self, nua_handle_t *nh)
+void
+tpsip_media_channel_attach_to_nua_handle (TpsipMediaChannel *self,
+                                          nua_handle_t *nh)
 {
   TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
 
@@ -1607,9 +1625,6 @@ priv_create_session (TpsipMediaChannel *channel,
 
   g_assert (priv->session == NULL);
 
-  /* Bind the channel object to the handle to handle NUA events */
-  priv_connect_nua_handlers (channel, nh);
-
   object_path = g_strdup_printf ("%s/MediaSession%u", priv->object_path, peer);
 
   DEBUG("allocating session, peer=%u", peer);
@@ -1691,6 +1706,10 @@ priv_outbound_call (TpsipMediaChannel *channel,
 
       nh = tpsip_conn_create_request_handle (priv->conn, peer);
       priv_create_session (channel, nh, peer);
+
+      /* Bind the channel object to the handle to handle NUA events */
+      tpsip_media_channel_attach_to_nua_handle (channel, nh);
+
       nua_handle_unref (nh);
     }
   else
@@ -1713,7 +1732,7 @@ _tpsip_media_channel_add_member (GObject *iface,
 
   if (priv->initiator == mixin->self_handle)
     {
-      TpIntSet *set;
+      TpIntSet *remote_pending;
 
       /* case a: an old-school outbound call
        * (we are the initiator, a new handle added with AddMembers) */
@@ -1723,21 +1742,20 @@ _tpsip_media_channel_add_member (GObject *iface,
       /* Backwards compatible behavior:
        * add the peer to remote pending without waiting for the actual request
        * to be sent */
-      set = tp_intset_new ();
-      tp_intset_add (set, handle);
+      remote_pending = tp_intset_new_containing (handle);
       tp_group_mixin_change_members (iface,
                                      "",
                                      NULL,    /* add */
                                      NULL,    /* remove */
                                      NULL,    /* local pending */
-                                     set,     /* remote pending */
+                                     remote_pending,     /* remote pending */
                                      mixin->self_handle,        /* actor */
                                      TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
-      tp_intset_destroy (set);
+      tp_intset_destroy (remote_pending);
 
-      /* update flags: allow removal and rescinding, no more adding */
-      tp_group_mixin_change_flags (iface,
-          TP_CHANNEL_GROUP_FLAG_CAN_REMOVE | TP_CHANNEL_GROUP_FLAG_CAN_RESCIND,
+      /* update flags: no more adding.
+       * Removal and rescinding are still allowed. */
+      tp_group_mixin_change_flags (iface, 0,
           TP_CHANNEL_GROUP_FLAG_CAN_ADD);
 
       return TRUE;
