@@ -1,7 +1,7 @@
 /*
  * sip-media-channel.h - Header for TpsipMediaChannel
  * Copyright (C) 2005 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2009 Nokia Corporation
  *
  * This work is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -76,8 +76,10 @@ gboolean _tpsip_media_channel_add_member (GObject *iface,
                                           const gchar *message,
                                           GError **error);
 
-void tpsip_media_channel_receive_invite (TpsipMediaChannel *self,
-                                         nua_handle_t *nh);
+void tpsip_media_channel_create_initial_streams (TpsipMediaChannel *self);
+
+void tpsip_media_channel_attach_to_nua_handle (TpsipMediaChannel *self,
+                                               nua_handle_t *nh);
 
 guint
 tpsip_media_channel_change_call_state (TpsipMediaChannel *self,
