@@ -328,8 +328,8 @@ static const gchar * const text_channel_allowed_properties[] = {
 };
 
 static void
-tpsip_text_factory_foreach_channel_class (TpChannelManager *manager,
-    TpChannelManagerChannelClassFunc func,
+tpsip_text_factory_type_foreach_channel_class (GType type,
+    TpChannelManagerTypeChannelClassFunc func,
     gpointer user_data)
 {
   GHashTable *table = g_hash_table_new_full (g_str_hash, g_str_equal,
@@ -346,7 +346,7 @@ tpsip_text_factory_foreach_channel_class (TpChannelManager *manager,
   g_hash_table_insert (table, (gchar *) text_channel_fixed_properties[1],
       value);
 
-  func (manager, table, text_channel_allowed_properties, user_data);
+  func (type, table, text_channel_allowed_properties, user_data);
 
   g_hash_table_destroy (table);
 }
@@ -586,7 +586,7 @@ tpsip_nua_i_message_cb (TpBaseConnection    *conn,
           handle, handle, NULL);
 
   tpsip_text_channel_receive (channel,
-      ev->nua, handle, text, len);
+      sip, handle, text, len);
 
   tp_handle_unref (contact_repo, handle);
 
@@ -635,7 +635,8 @@ channel_manager_iface_init (gpointer g_iface, gpointer iface_data)
   TpChannelManagerIface *iface = g_iface;
 
   iface->foreach_channel = tpsip_text_factory_foreach_channel;
-  iface->foreach_channel_class = tpsip_text_factory_foreach_channel_class;
+  iface->type_foreach_channel_class =
+    tpsip_text_factory_type_foreach_channel_class;
   iface->create_channel = tpsip_text_factory_create_channel;
   iface->request_channel = tpsip_text_factory_request_channel;
   iface->ensure_channel = tpsip_text_factory_ensure_channel;
