@@ -6,6 +6,8 @@ from sofiatest import go, exec_test
 import twisted.protocols.sip
 
 import dbus
+import email.utils
+import time
 import uuid
 
 # Test message channels
@@ -83,13 +85,10 @@ def test(q, bus, conn, sip):
 
     ua_via = twisted.protocols.sip.parseViaHeader(event.headers['via'][0])
 
-    q.expect('dbus-signal', signal='Sent')
-
     conn.ReleaseHandles(1, [handle])
 
-    url = twisted.protocols.sip.parseURL(self_uri)
-    msg = twisted.protocols.sip.Request('MESSAGE', url)
-    send_message(sip, ua_via, 'Hi')
+    call_id = 'XYZ@localhost'
+    send_message(sip, ua_via, 'Hi', call_id=call_id, time=1234567890)
 
     incoming_obj, handle = test_new_channel (q, bus, conn,
         target_uri=FROM_URL,
@@ -101,14 +100,20 @@ def test(q, bus, conn, sip):
     name = conn.InspectHandles(1, [handle])[0]
     assert name == FROM_URL
 
-    event = q.expect('dbus-signal', signal='Received')
-    assert event.args[5] == 'Hi'
+    event = q.expect('dbus-signal', signal='MessageReceived')
+    msg = event.args[0]
+    now = time.time()
+    assert msg[0]['message-token'] == "%s;cseq=%u" % (call_id, cseq_num)
+    assert now - 10 < msg[0]['message-received'] < now + 10
+    assert msg[0]['message-sent'] == 1234567890
+    assert msg[1]['content-type'] == 'text/plain'
+    assert msg[1]['content'] == 'Hi'
 
     # FIXME: times out for some reason, the response is in fact sent;
     # race condition with the earlier wait for 'dbus-signal'?
     #event = q.expect('sip-response', code=200)
 
-    iface.AcknowledgePendingMessages([event.args[0]])
+    iface.AcknowledgePendingMessages([msg[0]['pending-message-id']])
 
     # Test conversion from an 8-bit encoding.
     # Due to limited set of encodings available in some environments,
@@ -137,8 +142,7 @@ def test(q, bus, conn, sip):
 
     send_message(sip, ua_via, 'How are you doing now, old pal?',
                  sender=contact)
-    event = q.expect('dbus-signal', signal='Received')
-    assert tp_path_prefix + event.path == chan, (event.path, chan)
+    event = q.expect('dbus-signal', signal='Received', path=chan)
     assert event.args[5] == 'How are you doing now, old pal?'
     pending_msgs.append(tuple(event.args))
 
@@ -152,8 +156,7 @@ def test(q, bus, conn, sip):
     dbus.Interface(requested_obj, CHANNEL).Close()
     del requested_obj
 
-    event = q.expect('dbus-signal', signal='Closed')
-    assert tp_path_prefix + event.path == chan, (event.path, chan)
+    event = q.expect('dbus-signal', signal='Closed', path=chan)
 
     requested_obj, handle = test_new_channel (q, bus, conn,
         target_uri=contact,
@@ -162,14 +165,16 @@ def test(q, bus, conn, sip):
 
     # Expect Channel_Text_Message_Flag_Resqued to be set
     pending_msgs = [message_with_resqued(msg) for msg in pending_msgs]
+    # The first message is the delivery report of the message we failed to
+    # send so we'll skip it.
 
     iface = dbus.Interface(requested_obj, TEXT_TYPE)
 
     pending_res = iface.ListPendingMessages(False)
-    assert pending_msgs == pending_res, (pending_msgs, unwrap(pending_res))
+    assert pending_msgs == pending_res[1:], (pending_msgs, unwrap(pending_res)[1:])
 
     pending_res = iface.ListPendingMessages(True)
-    assert pending_msgs == pending_res, (pending_msgs, unwrap(pending_res))
+    assert pending_msgs == pending_res[1:], (pending_msgs, unwrap(pending_res)[1:])
 
     # There should be no pending messages any more
     pending_res = iface.ListPendingMessages(False)
@@ -197,7 +202,8 @@ def test(q, bus, conn, sip):
     q.expect('dbus-signal', signal='StatusChanged', args=[2,1])
 
 cseq_num = 1
-def send_message(sip, destVia, body, encoding=None, sender=FROM_URL):
+def send_message(sip, destVia, body,
+                 encoding=None, sender=FROM_URL, call_id=None, time=None):
     global cseq_num
     cseq_num += 1
     url = twisted.protocols.sip.parseURL('sip:testacc@127.0.0.1')
@@ -212,7 +218,9 @@ def send_message(sip, destVia, body, encoding=None, sender=FROM_URL):
     else:
         msg.addHeader('content-type', 'text/plain; charset=%s' % encoding)
     msg.addHeader('content-length', '%d' % len(msg.body))
-    msg.addHeader('call-id', uuid.uuid4().hex)
+    msg.addHeader('call-id', call_id or uuid.uuid4().hex)
+    if time is not None:
+        msg.addHeader('date', email.utils.formatdate(time, False, True))
     via = sip.getVia()
     via.branch = 'z9hG4bKXYZ'
     msg.addHeader('via', via.toString())
