@@ -27,7 +27,7 @@ G_BEGIN_DECLS
 
 gchar * tpsip_quote_string (const gchar *src);
 
-gchar * tpsip_unquote_string (const gchar *src, gsize len);
+gchar * tpsip_unquote_string (const gchar *src, gssize len);
 
 void tpsip_string_append_quoted (GString *buf, const gchar *text);
 
